---@type Tinkr
local Tinkr,
---@class Bastion
Bastion = ...

-- APL (Attack priority list) class
---@class Bastion.APL
---@field apl Bastion.APLActor[]
---@field variables table<string, any>
---@field last { attempt: Bastion.APLActor, success: Bastion.APLActor, time: number, index: number }
local APL = {}
APL.__index = APL

-- Constructor
---@param name string
---@return Bastion.APL
function APL:New(name)
    ---@class Bastion.APL
    local self = setmetatable({
        active = false,
        apl = {},
        variables = {},
        name = name,
        last = {
            success = {},
            attempt = {},
            time = -1,
            index = -1,
        },
    }, APL)

    return self
end

-- Add a variable to the APL
---@param name string
---@param value any
function APL:SetVariable(name, value)
    self.variables[name] = value
end

-- Get and evaluate a variable
---@param name string
---@return boolean
function APL:GetVariable(name)
    return self.variables[name]
end

---@class Bastion.APL.Actor.Variable.Table : Bastion.APLActor.Table.Base
---@field variable string
---@field cb fun(...):any

-- Add variable
---@param name string
---@param cb fun(...):any
---@return Bastion.APLActor
function APL:AddVariable(name, cb)
    local actor = Bastion.APLActor:New({
        type = "variable",
        variable = name,
        cb = cb,
        parentAPL = self,
    })
    table.insert(self.apl, actor)
    return actor
end

---@class Bastion.APL.Actor.Action.Table : Bastion.APLActor.Table.Base
---@field action string
---@field cb fun(...):any
---@field args? any[]

-- Add a manual action to the APL
---@generic A
---@param action string
---@param cb fun(...: A):any
---@param ... A
---@return Bastion.APLActor
function APL:AddAction(action, cb, ...)
    local actor = Bastion.APLActor:New({
        type = "action",
        action = action,
        cb = cb,
        parentAPL = self,
        args = SafePack(...),
    })
    table.insert(self.apl, actor)
    return actor
end

---@class Bastion.APL.Actor.Spell.Table : Bastion.APLActor.Table.Base
---@field spell Bastion.Spell
---@field condition? string|fun(self: Bastion.Spell, target: Bastion.Unit|false): boolean
---@field castableFunc false | fun(self: Bastion.Spell): boolean
---@field target Bastion.Unit | false
---@field onCastFunc false | fun(self: Bastion.Spell):any
---@field targetIf fun(self: Bastion.Spell): (Bastion.Unit) | false

-- Add a spell to the APL
---@param spell Bastion.Spell
---@param condition? string|fun(self: Bastion.Spell, target: Bastion.Unit|false):boolean
---@param targetIf? fun(self: Bastion.Spell): Bastion.Unit
---@return Bastion.APLActor
function APL:AddSpell(spell, condition, targetIf)
    local castableFunc = spell.CastableIfFunc
    local onCastFunc = spell.OnCastFunc
    local target = spell:GetTarget()

    local actor = Bastion.APLActor:New({
        type = "spell",
        spell = spell,
        condition = condition,
        castableFunc = castableFunc,
        target = target,
        onCastFunc = onCastFunc,
        parentAPL = self,
        targetIf = targetIf or false,
        executor = function(actor)
            local actorTable = actor.actor --[[@as Bastion.APL.Actor.Spell.Table]]
            local spellTarget = actorTable.targetIf and actorTable.targetIf(actorTable.spell) or actorTable.target
            return (not spellTarget or spellTarget:Exists())
                and (not actorTable.condition or actorTable.condition(actorTable.spell, spellTarget))
                and actorTable.spell:CastableIf(actorTable.castableFunc):OnCast(actorTable.onCastFunc):Cast(spellTarget)
        end
    })

    table.insert(self.apl, actor)

    return actor
end

---@class Bastion.APL.Actor.Item.Table : Bastion.APLActor.Table.Base
---@field item Bastion.Item
---@field condition? string | fun(self: Bastion.Item): boolean
---@field usableFunc false | fun(...): boolean
---@field target Bastion.Unit | nil

-- Add an item to the APL
---@param item Bastion.Item
---@param condition? string | fun(self: Bastion.Item): boolean
---@return Bastion.APLActor
function APL:AddItem(item, condition)
    local usableFunc = item.UsableIfFunc
    local target = item:GetTarget()

    local actor = Bastion.APLActor:New({
        type = "item",
        item = item,
        condition = condition,
        usableFunc = usableFunc,
        target = target,
        parentAPL = self,
    })

    table.insert(self.apl, actor)

    return actor
end

---@class Bastion.APL.Actor.APL.Table : Bastion.APLActor.Table.Base
---@field apl Bastion.APL
---@field condition? fun(...):boolean

-- Add an APL to the APL (for sub APLs)
---@param apl Bastion.APL
---@param condition? fun(...):boolean
---@return Bastion.APLActor
function APL:AddAPL(apl, condition)
    local actor = Bastion.APLActor:New({
        type = "apl",
        apl = apl,
        condition = condition,
        parentAPL = self,
    })
    table.insert(self.apl, actor)
    return actor
end

---@param name string
---@param enabled? boolean
function APL:ToggleAPL(name, enabled)
    for _, actor in ipairs(self.apl) do
        if actor.actor.type == "apl" and actor.name == name then
            actor.enabled = type(enabled) == "boolean" and enabled or not actor.enabled
            break
        end
    end
end

---@param actor Bastion.APLActor
function APL:UpdateLastAttempted(actor)
    self.active = true
    self.last.time = GetTime()
    self.last.attempt = actor
    self.last.index = actor.index
end

function APL:UpdateLastSuccessful()
    self.last.success = self.last.attempt
end

function APL:GetCurrentActor()
    if self.active then
        return self.last.attempt
    else
        return false
    end
end

---@param active boolean
function APL:SetActive(active)
    self.active = active
end

-- Execute the APL
function APL:Execute()
    for i, actor in ipairs(self.apl) do
        self:UpdateLastAttempted(actor)
        if actor.enabled and (not actor:HasTraits() or actor:Evaluate()) and actor:Execute() then
            self:UpdateLastSuccessful()
            return true
        end
    end
    self.active = false
    return false
end

---@class Bastion.APL.Actor.Sequencer.Table : Bastion.APLActor.Table.Base
---@field sequencer Bastion.Sequencer
---@field condition? fun(...):boolean

-- Add a Sequencer to the APL
---@param sequencer Bastion.Sequencer
---@param condition fun(...):boolean
---@return Bastion.APLActor
function APL:AddSequence(sequencer, condition)
    local actor = Bastion.APLActor:New({
        type = "sequencer",
        sequencer = sequencer,
        condition = condition,
        parentAPL = self,
    })
    table.insert(self.apl, actor)
    return actor
end

-- tostring
---@return string
function APL:__tostring()
    return string.format("APL(%s)", self.name)
end

Bastion.APL = APL
