---@type Tinkr
local Tinkr = ...

---@class Bastion.Globals.SpellName : { [spellId]: string }

---@class Bastion
local Bastion = {
    CombatEvents = {},
    Enabled = false,
    Interval = 0.15,
    Globals = {
        ---@type Bastion.Globals.SpellName
        SpellName = {}
    },
    ---@class Bastion.LoadedFiles.Table
    ---@field [number] { originalPath: string, loadedPath: string, reloadable: boolean, order: number, newPath: string }
    LoadedFiles = {},
    LoadThird = true,
    Paths = {
        Scripts = "scripts",
        BastionBase = "scripts/bastion",
        BastionScripts = "scripts/bastion/scripts",
        ThirdParty = "scripts/BastionScripts",
    },
}

function Bastion:__index(key)
    if Bastion[key] then
        return Bastion[key]
    end
    return rawget(self, key)
end

local bastionFiles = {
    "~/src/ClassMagic/ClassMagic",
    "~/src/List/List",
    "~/src/Command/Command",
    "~/src/Util/Util",
    "~/src/Library/Library",
    "~/src/Notification/Notification",
    "~/src/NotificationList/NotificationList",
    "~/src/Vector3/Vector3",
    "~/src/Sequencer/Sequencer",
    "~/src/Cache/Cache",
    "~/src/Cacheable/Cacheable",
    "~/src/Refreshable/Refreshable",
    "~/src/EventManager/EventManager",
    "~/src/Unit/Unit",
    "~/src/Aura/Aura",
    "~/src/APLTrait/APLTrait",
    "~/src/APLActor/APLActor",
    "~/src/APL/APL",
    "~/src/Module/Module",
    "~/src/UnitManager/UnitManager",
    "~/src/ObjectManager/ObjectManager",
    "~/src/Missile/Missile",
    "~/src/MissileManager/MissileManager",
    "~/src/Spell/Spell",
    "~/src/SpellBook/SpellBook",
    "~/src/Item/Item",
    "~/src/ItemBook/ItemBook",
    "~/src/AuraTable/AuraTable",
    "~/src/Class/Class",
    "~/src/Timer/Timer",
    "~/src/MythicPlusUtils/MythicPlusUtils",
    "~/src/Config/Config",
    "~/src/TimeToDie/TimeToDie",
}

---@param filePath string
function Bastion:CheckIfLoaded(filePath)
    for i, file in ipairs(Bastion.LoadedFiles) do
        if file.loadedPath == filePath or file.originalPath == filePath or file.newPath == filePath then
            return true
        end
    end
    return false
end

---@param path string
---@param extension string
---@return string
local function AppendExtension(path, extension)
    return string.format("%s.%s", path, extension)
end

---@param path string
---@param extensions string|string[]
local function CheckFileExtensions(path, extensions)
    local exts = {}
    if type(extensions) == "string" then
        exts = { extensions }
    else
        exts = extensions
    end

    for i, extension in ipairs(exts) do
        local newPath = path
        if newPath:sub(extension:len() * -1) ~= extension then
            newPath = AppendExtension(newPath, extension)
        end
        if FileExists(newPath) then
            return newPath:sub(1, (extension:len() + 2) * -1), true
        end
    end
    return path, false
end

--- 0 = Failed, 1 = Success, 2 = Already Loaded
---@param filePath string | { filePath: string, reloadable: boolean }
---@param ... any
---@return 0|1|2, table
function Bastion:Require(filePath, ...)
    local loadedFile = {
        originalPath = filePath.filePath or filePath,
        newPath = filePath.filePath or filePath,
        loadedPath = filePath.filePath or filePath,
        reloadable = not filePath.reloadable and false or false,
        order = #Bastion.LoadedFiles + 1,
    }

    local pathResolutionShortcut = loadedFile.originalPath:sub(1, 1) == "@" and Bastion.Paths.BastionScripts or
        loadedFile.originalPath:sub(1, 1) == "~" and Bastion.Paths.BastionBase or
        loadedFile.originalPath:sub(1, 1) == "!" and Bastion.Paths.Scripts

    if pathResolutionShortcut then
        loadedFile.newPath = string.format("%s%s", pathResolutionShortcut, loadedFile.originalPath:sub(2))
        loadedFile.loadedPath = loadedFile.newPath
    end

    if DirectoryExists(loadedFile.newPath) then
        local fileList = ListFiles(loadedFile.newPath)
        if #fileList > 0 then
            local loadResults = {}
            for i = 1, #fileList do
                local file = loadedFile.newPath .. fileList[i]
                local status, returns = Bastion:Require(file, ...)
                table.insert(loadResults, { file = file, status = status, returns = returns })
            end
            return 1, loadResults
        end
        --Log(string.format("Bastion:Require - No files found in directory: %s", loadedFile.newPath))
        return 0, SafePack(nil)
    end

    local found = false
    -- Check if file path has a .lua or .luac extension. If not, try to add one and check if the file exists
    loadedFile.loadedPath, found = CheckFileExtensions(loadedFile.newPath, { "lua", "luac" })

    if not found then
        --Log(string.format("Bastion:Require - Not Found: %s (%s)", loadedFile.newPath, loadedFile.originalPath))
        return 0, SafePack(nil)
    end

    if not loadedFile.reloadable then
        if Bastion:CheckIfLoaded(loadedFile.loadedPath) then
            --Log(string.format("Bastion:Require - Already loaded: %s (%s)", loadedFile.newPath, loadedFile.originalPath))
            return 2, SafePack(nil)
        end
    end

    table.insert(Bastion.LoadedFiles, loadedFile)
    return 1, SafePack(require(loadedFile.loadedPath, Bastion, ...))
end

---@param load? boolean
---@param folder? string
local function LoadThird(load, folder)
    load = type(load) == "nil" and Bastion.LoadThird or load
    folder = folder or Bastion.Paths.ThirdParty
    if not load then
        return
    end
    local thirdPartyModulesFolders = ListFolders(folder)
    for i = 1, #thirdPartyModulesFolders do
        local moduleFolder = thirdPartyModulesFolders[i]
        local currentFolderDir = string.format("%s/%s", folder, moduleFolder)
        local loaderFilePath = string.format("%s/%s", currentFolderDir, "loader")
        if FileExists(loaderFilePath .. ".lua") or FileExists(loaderFilePath .. ".luac") then
            Bastion:Require(loaderFilePath, currentFolderDir)
        end
    end
end

---@param toggle? boolean
function Bastion:Toggle(toggle)
    Bastion.Enabled = type(toggle) ~= "nil" and toggle or not Bastion.Enabled
end

local combatEventSuffix = {
    ["_DAMAGE"] = true,
    ["_MISSED"] = true,
    ["_ABSORBED"] = true,
    ["_DRAIN"] = true,
    ["_LEECH"] = true,
    ["_INTERRUPT"] = true,
    ["_DISPEL"] = true,
    ["_DISPEL_FAILED"] = true,
    ["_STOLEN"] = true,
    ["_EXTRA_ATTACKS"] = true,
    ["_AURA_APPLIED"] = true,
    ["_AURA_REMOVED"] = true,
    ["_AURA_APPLIED_DOSE"] = true,
    ["_AURA_REMOVED_DOSE"] = true,
    ["_AURA_REFRESH"] = true,
    ["_AURA_BROKEN"] = true,
    ["_AURA_BROKEN_SPELL"] = true,
    ["_CAST_START"] = true,
    ["_CAST_SUCCESS"] = true,
    ["_CAST_FAILED"] = true,
    ["_INSTAKILL"] = true,
    ["_DURABILITY_DAMAGE"] = true,
    ["_DURABILITY_DAMAGE_ALL"] = true,
    ["_CREATE"] = true,
    ["_SUMMON"] = true,
    ["_RESURRECT"] = true,
    ["_EMPOWER_START"] = true,
    ["_EMPOWER_END"] = true,
    ["_EMPOWER_INTERRUPT"] = true,
}

local combatEventPrefix = {
    "SWING",
    "RANGE",
    "SPELL",
    "SPELL_PERIODIC",
    "SPELL_BUILDING",
    "ENVIRONMENTAL"
}

local loaded = false
function Bastion:Load()
    if loaded then
        return self
    end
    for suffix, _ in pairs(combatEventSuffix) do
        for _, prefix in pairs(combatEventPrefix) do
            Bastion.CombatEvents[prefix .. suffix] = true
        end
    end

    for i = 1, #bastionFiles do
        self:Require(bastionFiles[i])
    end

    self.Globals.EventManager:RegisterWoWEvent("PLAYER_ENTERING_WORLD", function()
        self.UnitManager:ResetObjects()
    end)

    self.Globals.Command:Register('toggle', 'Toggle bastion on/off', function()
        self:Toggle()
        if self.Enabled then
            self.Util:Print("Enabled")
        else
            self.Util:Print("Disabled")
        end
    end)

    self.Globals.CombatTimer = self.Timer:New("combat", function()
        return UnitAffectingCombat("player")
    end)

    ---@param unitTarget UnitId
    self.Globals.EventManager:RegisterWoWEvent("UNIT_HEALTH", function(unitTarget)
        --Bastion.UnitManager:Get(unitTarget):UpdateHealth()
    end)

    ---@param unit UnitToken
    ---@param auras UnitAuraUpdateInfo
    self.Globals.EventManager:RegisterWoWEvent("UNIT_AURA", function(unit, auras)
        local u = self.UnitManager:Get(unit)

        if u:Exists() then
            u:GetAuras():OnUpdate(auras)
        end
    end)

    self.Globals.EventManager:RegisterWoWEvent("UNIT_SPELLCAST_SUCCEEDED", function(...)
        ---@type UnitIds, string, spellId
        local unit, castGUID, spellID = ...
        local spell = self.Globals.SpellBook:GetIfRegistered(spellID)

        if unit == "player" and spell then
            spell.lastCastAt = GetTime()

            if spell:GetPostCastFunction() then
                spell:GetPostCastFunction()(spell)
            end
        end
    end)

    local playerGuid = UnitGUID("player")
    local missed = {}

    Bastion.Globals.EventManager:RegisterWoWEvent("UNIT_COMBAT",
        ---@param unitTarget UnitIds
        ---@param event string
        ---@param flagText string
        ---@param amount number
        ---@param schoolMask number
        function(unitTarget, event, flagText, amount, schoolMask)
            --[[ local unit = Bastion.UnitManager:Get(unitTarget)
            if unit:IsAffectingCombat() then
                unit:SetLastCombatTime()
            end ]]
        end)

    Bastion.Globals.EventManager:RegisterWoWEvent("COMBAT_LOG_EVENT_UNFILTERED", function()
        local args = { CombatLogGetCurrentEventInfo() }

        ---@type string
        local subEvent = args[2]
        ---@type string
        local sourceGUID = args[4]
        ---@type string
        local destGUID = args[8]
        ---@type number
        local spellID = args[12]
        ---@type string
        local spellName = args[13]

        if subEvent == "SPELL_CAST_SUCCESS" then
            if not Bastion.Globals.SpellName[spellID] then
                Bastion.Globals.SpellName[spellID] = spellName
            end
        end

        local sourceUnit = Bastion.UnitManager:GetObject(sourceGUID)

        local destUnit = Bastion.UnitManager:GetObject(destGUID)

        --local t = GetTime()

        if sourceUnit and sourceUnit:IsAffectingCombat() then
            sourceUnit:SetLastCombatTime()
        end

        local updateDestUnit = destUnit and destUnit:IsAffectingCombat()

        if Bastion.CombatEvents[subEvent] and not updateDestUnit or ((sourceUnit and sourceUnit:IsValid() and destUnit and destUnit:IsValid()) and not UnitThreatSituation(sourceUnit.unit:unit() --[[@as string]], destUnit.unit:unit() --[[@as string]])) then
            updateDestUnit = true
            --[[ for key, val in pairs(combatEvents) do
                if val and subEvent:find(key) then
                end
            end ]]
        end

        if destUnit and updateDestUnit then
            destUnit:SetLastCombatTime()
        end

        --[[ if destUnit then
            if subEvent == "SPELL_MISSED" and sourceGUID == playerGuid and spellID == 408 then
                local missType = args[15]

                if missType == "IMMUNE" then
                    local castingSpell = sourceUnit:GetCastingOrChannelingSpell()

                    if castingSpell and type(castingSpell) == "table" then
                        if not missed[castingSpell:GetID()] then
                            missed[castingSpell:GetID()] = true
                        end
                    end
                end
            end
        end ]]
    end)

    self.Ticker = C_Timer.NewTicker(self.Interval, function()
        self.Globals.CombatTimer:Check()

        if self.Enabled then
            self.MissileManager:Refresh()
            self.ObjectManager:Refresh()
            self.TimeToDie:Refresh()
            self:TickModules()
        else
            self.UnitManager:ResetObjects()
        end
    end)

    self.Globals.Command:Register("dumpspells", "Dump spells to a file", function()
        local i = 1
        local rand = math.random(100000, 999999)
        while true do
            local spellName, spellSubName = GetSpellBookItemName(i, BOOKTYPE_SPELL)
            if not spellName then
                do
                    break
                end
            end

            -- use spellName and spellSubName here
            local spellID = select(7, GetSpellInfo(spellName))

            if spellID then
                spellName = spellName:gsub("[%W%s]", "")
                WriteFile("bastion-" .. UnitClass("player") .. "-" .. rand .. ".lua",
                    "local " .. spellName .. " = Bastion.Globals.SpellBook:GetSpell(" .. spellID .. ")\n", true)
            end
            i = i + 1
        end
    end)

    self.Globals.Command:Register("missed", "Dump the list of immune kidney shot spells", function()
        for k, v in pairs(missed) do
            self.Util:Print(k)
        end
    end)

    self:Require("@/Libraries/")
    self:Require("@/Modules/")
    self:Require("@/")
    LoadThird()
    loaded = true

    return self
end

return Bastion
