---@type Tinkr
local Tinkr,
---@class Bastion
Bastion = ...

---@class Bastion.Item.Traits.Use
---@field moving? boolean
---@field dead? boolean
---@field casting? boolean
---@field channeling? boolean
---@field byId? boolean

---@class Bastion.Item.Traits.Target
---@field exists? boolean
---@field player? boolean

-- Create a new Item class
---@class Bastion.Item
---@field itemID number
---@field UsableIfFunc boolean | fun(self:Bastion.Item):boolean
---@field PreUseFunc boolean | fun(self:Bastion.Item)
---@field target Bastion.Unit | false
---@field conditions table<string, { func: fun(self:Bastion.Item):boolean }>
---@field OnUseFunc boolean | fun(self:Bastion.Item)
---@field spellID number | nil
---@field lastUpdated boolean|number
---@field playerUsable boolean
---@field shouldUpdate boolean
local Item = {
    UsableIfFunc = false,
    PreUseFunc = false,
    OnUseFunc = false,
    wasLooking = false,
    lastUpdated = false,
    lastUseAttempt = 0,
    conditions = {},
}

local usableExcludes = {
    [18562] = true,
}

---@param itemId number | string
---@return number charges, number maxCharges, number start, number duration
local GetItemCharges = function(itemId)
    local _, spellId = C_Item.GetItemSpell(itemId)
    local charges, maxCharges, start, duration, chargeModRate = GetSpellCharges(spellId)
    return charges, maxCharges, start, duration
end

function Item:__index(k)
    local response = Bastion.ClassMagic:Resolve(Item, k)
    if response == nil then
        response = rawget(self, k)
    end

    if response == nil then
        error("Item:__index: " .. k .. " does not exist")
    end

    return response
end

-- Equals
---@param other Bastion.Item
---@return boolean
function Item:__eq(other)
    return self:GetID() == other:GetID()
end

-- tostring
---@return string
function Item:__tostring()
    return "Bastion.__Item(" .. self:GetID() .. ")" .. " - " .. self:GetName()
end

-- Constructor
---@param id number
function Item:New(id)
    ---@class Bastion.Item
    local self = setmetatable({}, Item)
    self.itemID = id
    self.spell = false
    self.playerUsable = C_PlayerInfo.CanUseItem(id)
    self.shouldUpdate = false
    self.traits = {
        use = {
            moving = true,
            dead = false,
            casting = false,
            channeling = false,
            byId = false,
        },
        target = {
            exists = true,
            player = false,
        }
    }

    -- C_PlayerInfo.CanUseItem

    if not C_Item.IsItemDataCachedByID(id) then
        self.shouldUpdate = true
        C_Item.RequestLoadItemDataByID(id)
    else
        local name, spellID = GetItemSpell(self:GetID())
        if spellID then
            self.spellID = spellID
            Bastion.Globals.SpellBook:GetSpell(spellID)
        end
    end

    return self
end

function Item:ShouldUpdate()
    if self.shouldUpdate and not self.lastUpdated or GetTime() - self.lastUpdated > 30 then
        return true
    end
end

---@class Bastion.Item.Traits.Params
---@field use? Bastion.Item.Traits.Use
---@field target? Bastion.Item.Traits.Target

---@param traits Bastion.Item.Traits.Params
function Item:SetTraits(traits)
    for _, itemTrait in pairs({ "use", "target" }) do
        if type(traits[itemTrait]) == "table" then
            local currentTrait = traits[itemTrait]
            for traitKey, traitValue in pairs(self.traits[itemTrait]) do
                if type(currentTrait[traitKey]) ~= "nil" then
                    self.traits[itemTrait][traitKey] = currentTrait[traitKey]
                end
            end
        end
    end
    return self
end

function Item:EvaluateTraits()
    local player = Bastion.UnitManager:Get("player")

    if not self.traits.use.moving and player:IsMoving() then
        return false
    end

    if not self.traits.use.dead and player:IsDead() then
        return false
    end

    if not self.traits.use.casting and player:IsCasting() then
        return false
    end

    if not self.traits.use.channeling and player:IsChanneling() then
        return false
    end

    if self.traits.target.exists then
        local target = self:GetTarget()
        if type(target) ~= "boolean" and not target:Exists() or target == false then
            return false
        end
    end

    return true
end

function Item:Update()
    if self:ShouldUpdate() then
        if not self.spellID and self.playerUsable then
            local name, spellID = GetItemSpell(self:GetID())
            if spellID then
                self.spellID = spellID
                self.shouldUpdate = false
                Bastion.Globals.SpellBook:GetSpell(spellID)
            end
        end
    end
end

-- Get the Items id
---@return number
function Item:GetID()
    return self.itemID
end

-- Get the Items name
---@return string
function Item:GetName()
    return select(1, GetItemInfo(self:GetID()))
end

-- Get the Items icon
---@return number
function Item:GetIcon()
    return select(3, GetItemInfo(self:GetID()))
end

-- Get the Items cooldown
---@return number
function Item:GetCooldown()
    return select(2, C_Container.GetItemCooldown(self:GetID()))
end

-- Return the Usable function
---@return function | boolean
function Item:GetUsableFunction()
    return self.UsableIfFunc
end

-- Return the preUse function
---@return function | boolean
function Item:GetPreUseFunction()
    return self.PreUseFunc
end

-- Get the on Use func
---@return function | boolean
function Item:GetOnUseFunction()
    return self.OnUseFunc
end

-- Get the Items cooldown remaining
---@return number
function Item:GetCooldownRemaining()
    local start, duration = C_Container.GetItemCooldown(self:GetID())
    return start == 0 and 0 or start + duration - GetTime()
end

-- Use the Item
---@param unit? Bastion.Unit
function Item:UseByID(unit)
    local target = unit and unit:GetOMToken() or self.traits.target.player and "player" or "target"

    RunMacroText(string.format("/use [@%s] item:%d", target, self:GetID()))

    Bastion.Util:Debug("Using by id", self)
end

-- Use the Item
---@param unit? Bastion.Unit
---@param condition? string | fun(self:Bastion.Item):boolean
---@return boolean
function Item:Use(unit, condition)
    if not self:Usable() then
        return false
    end
    if condition then
        if type(condition) == "string" and not self:EvaluateCondition(condition) then
            return false
        elseif type(condition) == "function" and not condition(self) then
            return false
        end
    end

    -- Call pre Use function
    if self:GetPreUseFunction() then
        self:GetPreUseFunction()(self)
    end

    local target = unit or self.traits.target.player and Bastion.UnitManager:Get("player") or
        self.traits.target.exists and self:TargetExists() and self:GetTarget() or false
    if not target then
        return false
    end
    -- Check if the mouse was looking
    self.wasLooking = IsMouselooking() --[[@as boolean]]
    UseItemByName(self.traits.use.byId and self:GetID() or self:GetName(), target:GetOMToken())
    Bastion.Util:Debug("Using", self)

    -- Set the last Use time
    self.lastUseAttempt = GetTime()

    -- Call post Use function
    if self:GetOnUseFunction() then
        self:GetOnUseFunction()(self)
    end

    return true
end

-- Last use attempt
---@return number
function Item:GetLastUseAttempt()
    return self.lastUseAttempt
end

-- Time since last attepmt
---@return number
function Item:GetTimeSinceLastUseAttempt()
    return GetTime() - self:GetLastUseAttempt()
end

-- Check if the Item is known
function Item:IsEquipped()
    return IsEquippedItem(self:GetID())
end

function Item:GetEquippedSlot()
    if not self:IsEquippable() and not self:IsEquipped() then
        return false
    end
    local itemId = self:GetID()
    local slots = {
        [1] = INVSLOT_AMMO,
        [2] = INVSLOT_HEAD,
        [3] = INVSLOT_NECK,
        [4] = INVSLOT_SHOULDER,
        [5] = INVSLOT_BODY,
        [6] = INVSLOT_CHEST,
        [7] = INVSLOT_WAIST,
        [8] = INVSLOT_LEGS,
        [9] = INVSLOT_FEET,
        [10] = INVSLOT_WRIST,
        [11] = INVSLOT_HAND,
        [12] = INVSLOT_FINGER1,
        [13] = INVSLOT_FINGER2,
        [14] = INVSLOT_TRINKET1,
        [15] = INVSLOT_TRINKET2,
        [16] = INVSLOT_BACK,
        [17] = INVSLOT_MAINHAND,
        [18] = INVSLOT_OFFHAND,
        [19] = INVSLOT_RANGED,
        [20] = INVSLOT_TABARD,
    }
    for i = 1, #slots do
        local slot = slots[i]
        if GetInventoryItemID("player", slot) == itemId then
            return slot
        end
    end
    return false
end

-- Check if the Item is on cooldown
---@return boolean
function Item:IsOnCooldown()
    local _, duration = C_Container.GetItemCooldown(self:GetID())
    return duration > 0
end

-- Check if the Item is usable
---@return boolean
function Item:IsUsable()
    local usable, noMana = IsUsableItem(self:GetID())
    return usable or usableExcludes[self:GetID()]
end

-- Check if the Item is Usable
---@return boolean
function Item:IsEquippedAndUsable()
    return ((self:IsEquippable() and self:IsEquipped()) or (not self:IsEquippable() and self:IsUsable()))
        and not self:IsOnCooldown()
end

-- Is equippable
function Item:IsEquippable()
    return IsEquippableItem(self:GetID())
end

-- Check if the Item is Usable
---@return boolean
function Item:Usable()
    if not self:EvaluateTraits() then
        return false
    end
    if self:GetUsableFunction() and not self:GetUsableFunction()(self) then
        return false
    end

    return self:IsEquippedAndUsable()
end

-- Set a script to check if the Item is Usable
---@param func fun(self:Bastion.Item):boolean
---@return Bastion.Item
function Item:UsableIf(func)
    self.UsableIfFunc = func
    return self
end

-- Set a script to run before the Item has been Use
---@param func fun(self:Bastion.Item)
---@return Bastion.Item
function Item:PreUse(func)
    self.PreUseFunc = func
    return self
end

-- Set a script to run after the Item has been Use
---@param func fun(self:Bastion.Item)
---@return Bastion.Item
function Item:OnUse(func)
    self.OnUseFunc = func
    return self
end

-- Get was looking
---@return boolean
function Item:GetWasLooking()
    return self.wasLooking
end

-- Click the Item
---@param x number
---@param y number
---@param z number
---@return boolean
function Item:Click(x, y, z)
    if type(x) == "table" then
        x, y, z = x.x, x.y, x.z
    end
    if IsSpellPending() == 64 then
        MouselookStop()
        Click(x, y, z)
        if self:GetWasLooking() then
            MouselookStart()
        end
        return true
    end
    return false
end

-- Check if the Item is Usable and Use it
---@param unit Bastion.Unit
---@return boolean
function Item:Call(unit)
    if self:Usable() then
        self:Use(unit)
        return true
    end
    return false
end

-- Check if the Item is in range of the unit
---@param unit Bastion.Unit
---@return boolean
function Item:IsInRange(unit)
    if not C_Item.ItemHasRange(self:GetID()) or not self.spellID then
        return true
    end

    local itemSpell = self:GetSpell()

    if itemSpell then
        return itemSpell:IsInRange(unit)
    end
    return false

    --[[ local name, rank, icon, UseTime, Itemmin, Itemmax, ItemID = GetItemInfo(self:GetID())

    local them = Object(unit:GetOMToken())

    local tx, ty, tz = ObjectPosition(unit:GetOMToken())
    local px, py, pz = ObjectPosition("player")

    if not them then
        return false
    end

    if tx == 0 and ty == 0 and tz == 0 then
        return true
    end

    local combatReach = ObjectCombatReach("player")
    local themCombatReach = ObjectCombatReach(unit:GetOMToken())

    if Bastion.UnitManager["player"]:InMelee(unit) and Itemmin == 0 then
        return true
    end

    local distance = FastDistance(px, py, pz, tx, ty, tz)

    if Itemmax and distance >= Itemmin and distance <= combatReach + themCombatReach + Itemmax then
        return true
    end

    return false ]]
end

-- Get the last use time
function Item:GetLastUseTime()
    return Bastion.Globals.SpellBook:GetSpell(self:GetID()):GetLastCastTime()
end

-- Get time since last use
---@return number
function Item:GetTimeSinceLastUse()
    if not self:GetLastUseTime() then
        return math.huge
    end
    return GetTime() - self:GetLastUseTime()
end

-- Get the Items charges
---@return number
function Item:GetCharges()
    return select(1, GetItemCharges(self:GetID()))
end

function Item:GetCount()
    return C_Item.GetItemCount(self:GetID(), false, false, false)
end

-- Get the Items charges remaining
---@return number
function Item:GetChargesRemaining()
    local charges, maxCharges, start, duration = GetItemCharges(self:GetID())
    return charges
end

-- Create a condition for the Item
---@param name string
---@param func fun(self:Bastion.Item)
---@return Bastion.Item
function Item:Condition(name, func)
    self.conditions[name] = {
        func = func,
    }
    return self
end

-- Get a condition for the Item
---@param name string
function Item:GetCondition(name)
    local condition = self.conditions[name]
    if condition then
        return condition
    end

    return nil
end

-- Evaluate a condition for the Item
---@param name string
---@return boolean
function Item:EvaluateCondition(name)
    local condition = self:GetCondition(name)
    if condition then
        return condition.func(self)
    end

    return false
end

-- Check if the Item has a condition
---@param name string
---@return boolean
function Item:HasCondition(name)
    local condition = self:GetCondition(name)
    if condition then
        return true
    end

    return false
end

-- Set the Items target
---@param unit Bastion.Unit
function Item:SetTarget(unit)
    self.target = unit
    return self
end

-- Get the Items target
function Item:GetTarget()
    return self.traits.target.player and Bastion.UnitManager:Get("player") or self.target
end

function Item:TargetExists()
    return self:GetTarget() and self:GetTarget():Exists() or false
end

-- IsMagicDispel
---@return boolean
function Item:IsMagicDispel()
    return ({
        [88423] = true,
    })[self:GetID()]
end

-- IsCurseDispel
---@return boolean
function Item:IsCurseDispel()
    return ({
        [88423] = true,
    })[self:GetID()]
end

-- IsPoisonDispel
---@return boolean
function Item:IsPoisonDispel()
    return ({
        [88423] = true,
    })[self:GetID()]
end

-- IsDiseaseDispel
---@return boolean
function Item:IsDiseaseDispel()
    return ({})[self:GetID()]
end

---@param item Bastion.Item
---@return boolean
function Item:IsItem(item)
    return self:GetID() == item:GetID()
end

-- Get the Items spell
---@return Bastion.Spell | nil
function Item:GetSpell()
    if self.spellID then
        return Bastion.Globals.SpellBook:GetSpell(self.spellID)
    end

    return nil
end

Bastion.Item = Item
